"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

import unittest
import numpy as np

# Import the core module
from wa_simulator.vehicle import WALinearKinematicBicycle

# -----
# Tests
# -----


class TestWALinearKinematicBicycle(unittest.TestCase):
    """Tests method related to the WALinearKinematicBicycle"""

    pass


if __name__ == '__main__':
    unittest.main()
